import os

import redis


if 'REDIS_URL' in os.environ:
    redis = redis.from_url(os.environ['REDIS_URL'])
else:
    redis = redis.Redis()


class Field(object):

    def __init__(self, initial):
        self.initial = initial

    def get(self, obj, key):
        ret = redis.get('tg:%s:%s' % (obj.chat_id, key))
        if ret is None:
            ret = self.initial
            self.set(obj, key, ret)
        return ret

    def set(self, obj, key, value):
        redis.set('tg:%s:%s' % (obj.chat_id, key), value)
        return value


class Integer(Field):

    def get(self, obj, key):
        return int(super().get(obj, key))

    def set(self, obj, key, value):
        super().set(obj, key, int(value))



from importlib import import_module
import asyncio
import os
import logging

import click

from tg.dispatcher import GameDispatcher


@click.command()
@click.argument('game')
@click.option('--token', default=os.environ.get('TELEGRAM_TOKEN'),
              help='Telegram Bot API token (get it from http://telegram.me/BotFather)')
@click.option('--debug', is_flag=True)
def main(game, token, debug):
    """Start a game bot server

    Pass a game module name to it, for example:

    > telegram-game tg.examples.quiz
    """
    logging.basicConfig(
        level=logging.DEBUG if debug else logging.INFO,
        format="%(asctime)s %(name)-16s %(levelname)-8s %(message)s"
    )
    game = import_module(game)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(GameDispatcher(game.Game).run())



import random

from tg.api import Chat
from tg.fields import Field


class BaseGame:
    """The base Game class

    Each player works with his own Game() instance.

    Use :py:mod:`tg.fields` to add a persistent attributes to Game, see
    py:mod:`tg.examples`.
    """

    def __init__(self, chat_id, queue, api):
        self.chat_id = chat_id
        self.queue = queue
        self.api = api
        self.chat = Chat(api, chat_id)

    def __getattribute__(self, key):
        value = super().__getattribute__(key)
        if isinstance(value, Field):
            return value.get(self, key)
        return value

    def __setattr__(self, key, value):
        if hasattr(self, key):
            attr = super().__getattribute__(key)
            if isinstance(attr, Field):
                attr.set(self, key, value)
                return
        super().__setattr__(key, value)

    @classmethod
    async def prepare(self, loop):
        """Prepare the game

        Use this method to prepare common game resources, establish the database
        connection, and so on.
        """
        pass

    async def send(self, msg, *args, **kwargs):
        if isinstance(msg, list):
            msg = random.choice(msg)
        return await self.chat.sendMessage(msg.format(*args, **kwargs))

    async def recv(self):
        """Get the next message from player."""
        return await self.queue.get()

    async def start(self):
        """Game logic implementation starts here.

        Use `await self.recv()` to get input from player, and `await
        self.chat.sendMessage()` to send a reply.
        """
        raise NotImplementedError()



"""Message dispatching logic"""

import asyncio
import aiohttp
import logging
import os

from tg.api import BotAPI


logger = logging.getLogger(__name__)


class GameDispatcher():

    def __init__(self, game_class, api_token=None, queue_maxsize=5, loop=None):

        self.game_class = game_class

        if api_token is None:
            api_token = os.environ['TELEGRAM_TOKEN']

        self.queue_maxsize = 5

        if loop is None:
            loop = asyncio.get_event_loop()

        self.loop = loop

        self.http = aiohttp.ClientSession(loop=loop)
        self.api = BotAPI(api_token, self.http)

        self.queues = {}

    async def run(self):
        await self.game_class.prepare(self.loop)
        info = await self.api.getMe()
        logger.info("Starting bot: %s", info)
        await self.poll_get_updates()

    async def poll_get_updates(self):
        while True:
            updates = await self.api.getUpdates(timeout=30.)
            self.loop.create_task(self.dispatch_updates(updates))

    async def dispatch_updates(self, updates):

        for update in updates:

            if 'message' in update:

                chat_id = update['message']['chat']['id']
                msg = update['message']
                logger.info('#%d @%s: %s', chat_id,
                            msg.get('from', {}).get('username'),
                            msg.get('text', 'NO_TEXT')[:60].replace('\n', '\\n'))

                if chat_id not in self.queues:
                    self.queues[chat_id] = asyncio.Queue(self.queue_maxsize)
                    game = self.game_class(chat_id, self.queues[chat_id], self.api)
                    self.loop.create_task(game.start())

                try:
                    self.queues[chat_id].put_nowait(update)
                except asyncio.QueueFull:
                    logger.warning('QueueFull for %s', chat_id)
                    self.loop.create_task(
                        self.sendMessage(
                            chat_id, 'Не так быстро!',
                            reply_to_message_id=update['message']['id']
                        )
                    )

            else:
                logger.warning("Unsupported update type: %s", update)



import logging
import random

from tg import fields, BaseGame, Messages


logger = logging.getLogger(__name__)


class M(Messages):
    PREAMBLE = "I'm thinking of a number from 1 to 10! Try to guess it!"
    LESS = [
        "You didn't guess right!",
        "You failed!",
        "Too high!",
    ]
    GREATER = [
        "Wrong!",
        "No.",
        "Too low!",
    ]
    BAD_INPUT = [
        "Are you cheating?!",
    ]
    SUCCESS = "You guessed! You guessed the number {0} times for {1} tries!"


class Game(BaseGame):

    guessed = fields.Integer(0)
    tries = fields.Integer(0)

    async def start(self):

        while True:

            num = random.randint(1, 10)

            await self.send(M.PREAMBLE)

            while True:

                msg = await self.recv()

                try:
                    guess = int(msg['message']['text'])
                except (KeyError, ValueError):
                    await self.send(M.BAD_INPUT)
                    continue

                self.tries += 1

                if guess == num:
                    self.guessed = self.guessed + 1
                    await self.send(M.SUCCESS, self.guessed, self.tries)
                    break
                elif guess < num:
                    await self.send(M.GREATER)
                else:
                    await self.send(M.LESS)



"""Telegram API implementation."""

import asyncio
import aiohttp
import logging

logger = logging.getLogger(__name__)


RETRY_TIMEOUT = 30
RETRY_CODES = [429, 500, 502, 503, 504]


class BotAPIError(Exception):
    pass


class BotAPI:
    """Telegram Bot API Wrapper.
    """

    API_URL = "https://api.telegram.org"

    def __init__(self, api_token, session=None):
        self.api_token = api_token
        self._http = session or aiohttp.ClientSession()
        self._offset = None

    async def __call__(self, method, **params):
        """Call Telegram Bot API.

        :param str method: Telegram API method
        :param params: Arguments for the method call
        """

        logger.debug("API request: %s %s", method, params)

        url = "{0}/bot{1}/{2}".format(self.API_URL, self.api_token, method)

        while True:

            async with self._http.post(url, data=params) as response:

                if response.status == 200:
                    data = await response.json()
                    if data['ok']:
                        return data['result']
                    else:
                        raise BotAPIError(data['description'])

                elif response.status in RETRY_CODES:
                    logger.info("Server returned %d, retrying in %d sec.",
                                response.status, RETRY_TIMEOUT)

                else:
                    if response.headers['content-type'] == 'application/json':
                        data = await response.json()
                        err_msg = data["description"]
                    else:
                        err_msg = await response.read()
                    raise BotAPIError(err_msg)

            await asyncio.sleep(RETRY_TIMEOUT)

    def getMe(self):
        return self('getMe')

    def sendMessage(self, chat_id, text, parse_mode=None,
                    disable_web_page_preview=None, disable_notification=None,
                    reply_to_message_id=None, reply_markup=None):
        """Use this method to send text messages.

        :param int|str chat_id:
            Unique identifier for the target chat or username of the target
            channel (in the format @channelusername)
        :param str text: Text of the message to be sent
        :param str parse_mode: Send Markdown or HTML, if you want Telegram apps
            to show bold, italic, fixed-width text or inline URLs in your bot's
            message.
        :param bool disable_web_page_preview: Disables link previews for links
            in this message
        :param bool disable_notification: Sends the message silently. iOS users
            will not receive a notification, Android users will receive a
            notification with no sound.
        :param int reply_to_message_id: If the message is a reply, ID of the
            original message
        :param str reply_markup: Additional interface options. A JSON-serialized
            object for an inline keyboard, custom reply keyboard, instructions
            to hide reply keyboard or to force a reply from the user.
            InlineKeyboardMarkup or ReplyKeyboardMarkup or ReplyKeyboardHide or
            ForceReply.

        :return: On success, the sent Message is returned.
        """
        params = {
            'chat_id': chat_id,
            'text': text,
        }
        if parse_mode is not None:
            params['parse_mode'] = parse_mode
        if disable_web_page_preview is not None:
            params['disable_web_page_preview'] = disable_web_page_preview
        if disable_notification is not None:
            params['disable_notification'] = disable_notification
        if reply_to_message_id is not None:
            params['reply_to_message_id'] = reply_to_message_id
        if reply_markup is not None:
            params['reply_markup'] = reply_markup
        return self('sendMessage', **params)

    async def getUpdates(self, offset=None, limit=None, timeout=None):
        """Use this method to receive incoming updates using long polling
        (`wiki`_).

        It automatically manage offset, so normaly you only have to pass a
        reasonable `timeout` value here (30 seconds is a good one, for example).

        :param int offset:

            Identifier of the first update to be returned. Must be greater by
            one than the highest among the identifiers of previously received
            updates. By default, updates starting with the earliest unconfirmed
            update are returned. An update is considered confirmed as soon as
            getUpdates is called with an _offset_ higher than its update_id.

            The negative offset can be specified to retrieve updates starting
            from -_offset_ update from the end of the updates queue. All
            previous updates will forgotten.

        :param int limit:

            Limits the number of updates to be retrieved. Values between 1—100
            are accepted. Defaults to 100.

        :param int timeout:

            Timeout in seconds for long polling. Defaults to 0, i.e. usual short
            polling

        :return: An Array of Update objects is returned.

        .. _wiki: http://en.wikipedia.org/wiki/Push_technology#Long_polling
        """

        params = {}

        if offset is None:
            offset = self._offset

        if offset is not None:
            params['offset'] = offset
        if limit is not None:
            params['limit'] = limit
        if timeout is not None:
            params['timeout'] = timeout

        updates = await self('getUpdates', **params)

        if updates:
            self._offset = max(i['update_id'] for i in updates) + 1

        return updates


class Chat:
    """An API wrapper for chat-related methods"""

    def __init__(self, api, chat_id):
        self.api = api
        self.chat_id = chat_id

    def sendMessage(self, text, parse_mode=None,
                    disable_web_page_preview=None, disable_notification=None,
                    reply_to_message_id=None, reply_markup=None):
        """Use this method to send text messages.

        :param str text: Text of the message to be sent
        :param str parse_mode: Send Markdown or HTML, if you want Telegram apps
            to show bold, italic, fixed-width text or inline URLs in your bot's
            message.
        :param bool disable_web_page_preview: Disables link previews for links
            in this message
        :param bool disable_notification: Sends the message silently. iOS users
            will not receive a notification, Android users will receive a
            notification with no sound.
        :param int reply_to_message_id: If the message is a reply, ID of the
            original message
        :param str reply_markup: Additional interface options. A JSON-serialized
            object for an inline keyboard, custom reply keyboard, instructions
            to hide reply keyboard or to force a reply from the user.
            InlineKeyboardMarkup or ReplyKeyboardMarkup or ReplyKeyboardHide or
            ForceReply.

        :return: On success, the sent Message is returned.
        """
        params = {
            'chat_id': self.chat_id,
            'text': text,
        }
        if parse_mode is not None:
            params['parse_mode'] = parse_mode
        if disable_web_page_preview is not None:
            params['disable_web_page_preview'] = disable_web_page_preview
        if disable_notification is not None:
            params['disable_notification'] = disable_notification
        if reply_to_message_id is not None:
            params['reply_to_message_id'] = reply_to_message_id
        if reply_markup is not None:
            params['reply_markup'] = reply_markup
        return self.api('sendMessage', **params)
